@@ -44,6 +44,8 @@
 #include <string.h>
 #include <typeinfo>
 
+using namespace CppyyLegacy;
+
 // temp
 #include <iostream>
 typedef CPyCppyy::Parameter Parameter;
@@ -205,7 +207,7 @@ class ApplicationStarter {
 public:
     ApplicationStarter() {
     // initialize ROOT early to guarantee proper order of shutdown later on (gROOT is a
-    // macro that resolves to the ROOT::GetROOT() function call)
+    // macro that resolves to the ::CppyyLegacy::GetROOT() function call)
         (void)gROOT;
 
     // setup dummy holders for global and std namespaces
@@ -221,6 +223,13 @@ public:
     // add a dummy global to refer to as null at index 0
         g_globalvars.push_back(nullptr);
 
+    // fill out the builtins
+        std::set<std::string> bi{g_builtins};
+        for (const auto& name : bi) {
+            for (const char* a : {"*", "&", "*&", "[]", "*[]"})
+                g_builtins.insert(name+a);
+        }
+
     // disable fast path if requested
         if (getenv("CPPYY_DISABLE_FASTPATH")) gEnableFastPath = false;
 
@@ -242,10 +251,6 @@ public:
                "#include <utility>";
         gInterpreter->ProcessLine(code);
 
-    // make sure we run in batch mode as far as ROOT graphics is concerned
-        if (!getenv("ROOTSYS"))
-            gROOT->SetBatch(kTRUE);
-
     // create helpers for comparing thingies
         gInterpreter->Declare(
             "namespace __cppyy_internal { template<class C1, class C2>"
@@ -254,26 +259,25 @@ public:
             "namespace __cppyy_internal { template<class C1, class C2>"
             " bool is_not_equal(const C1& c1, const C2& c2) { return (bool)(c1 != c2); } }");
 
+    // helper for multiple inheritance
+        gInterpreter->Declare("namespace __cppyy_internal { struct Sep; }");
+
     // retrieve all initial (ROOT) C++ names in the global scope to allow filtering later
-        if (!getenv("CPPYY_NO_ROOT_FILTER")) {
-            gROOT->GetListOfGlobals(true);             // force initialize
-            gROOT->GetListOfGlobalFunctions(true);     // id.
-            std::set<std::string> initial;
-            Cppyy::GetAllCppNames(GLOBAL_HANDLE, initial);
-            gInitialNames = initial;
+        gROOT->GetListOfGlobals(true);             // force initialize
+        gROOT->GetListOfGlobalFunctions(true);     // id.
+        std::set<std::string> initial;
+        Cppyy::GetAllCppNames(GLOBAL_HANDLE, initial);
+        gInitialNames = initial;
 
 #ifndef WIN32
-            gRootSOs.insert("libCore.so ");
-            gRootSOs.insert("libRIO.so ");
-            gRootSOs.insert("libThread.so ");
-            gRootSOs.insert("libMathCore.so ");
+        gRootSOs.insert("libCoreLegacy.so ");
+        gRootSOs.insert("libRIOLegacy.so ");
+        gRootSOs.insert("libThreadLegacy.so ");
 #else
-            gRootSOs.insert("libCore.dll ");
-            gRootSOs.insert("libRIO.dll ");
-            gRootSOs.insert("libThread.dll ");
-            gRootSOs.insert("libMathCore.dll ");
+        gRootSOs.insert("libCoreLegacy.dll ");
+        gRootSOs.insert("libRIOLegacy.dll ");
+        gRootSOs.insert("libThreadLegacy.dll ");
 #endif
-        }
 
     // start off with a reasonable size placeholder for wrappers
         gWrapperHolder.reserve(1024);
@@ -337,6 +341,13 @@ bool Cppyy::Compile(const std::string& code)
     return gInterpreter->Declare(code.c_str());
 }
 
+std::string Cppyy::ToString(TCppType_t klass, TCppObject_t obj)
+{
+    if (klass && obj && !IsNamespace((TCppScope_t)klass))
+        return gInterpreter->ToString(GetScopedFinalName(klass).c_str(), (void*)obj);
+    return "";
+}
+
 
 // name to opaque C++ scope representation -----------------------------------
 std::string Cppyy::ResolveName(const std::string& cppitem_name)
@@ -369,8 +380,7 @@ std::string Cppyy::ResolveName(const std::string& cppitem_name)
 
 // check data types list (accept only builtins as typedefs will
 // otherwise not be resolved)
-    TDataType* dt = gROOT->GetType(tclean.c_str());
-    if (dt && dt->GetType() != kOther_t) return dt->GetFullTypeName();
+    if (IsBuiltin(tclean)) return tclean;
 
 // special case for enums
     if (IsEnum(cppitem_name))
@@ -404,8 +414,11 @@ std::string Cppyy::ResolveName(const std::string& cppitem_name)
         return resolved;
     }
 
-// typedefs
-    return TClassEdit::ResolveTypedef(tclean.c_str(), true);
+// typedefs etc. (and cleanup STL classes for consistency)
+    tclean = TClassEdit::ResolveTypedef(tclean.c_str(), true);
+    if (tclean.compare(0, 6, "const ") != 0)
+        return TClassEdit::ShortType(tclean.c_str(), 2);
+    return "const " + TClassEdit::ShortType(tclean.c_str(), 2);
 }
 
 
@@ -448,7 +461,7 @@ std::string Cppyy::ResolveEnum(const std::string& enum_type)
     if (res != resolved_enum_types.end())
         return res->second;
 
-// desugar the type before resolving
+// remove qualifiers and desugar the type before resolving
     std::string et_short = TClassEdit::ShortType(enum_type.c_str(), 1);
     if (et_short.find("(anonymous") == std::string::npos) {
         TEnum* ee = nullptr;
@@ -458,7 +471,8 @@ std::string Cppyy::ResolveEnum(const std::string& enum_type)
             ee = ((TListOfEnums*)gROOT->GetListOfEnums())->GetObject(et_short.c_str());
         else {
             TClass* cl = TClass::GetClass(scope_name.c_str());
-            if (cl) ee = ((TListOfEnums*)cl->GetListOfEnums())->GetObject(et_short.c_str());
+            if (cl) ee = ((TListOfEnums*)cl->GetListOfEnums())->GetObject(
+                et_short.substr(scope_name.size()+2, std::string::npos).c_str());
         }
 
         if (ee) {
@@ -517,8 +531,6 @@ Cppyy::TCppScope_t Cppyy::GetScope(const std::string& sname)
         if (result) return result;
     }
 
-    if (result) return result;
-
 // use TClass directly, to enable auto-loading; class may be stubbed (eg. for
 // function returns) or forward declared, leading to a non-null TClass that is
 // otherwise invalid/unusable
@@ -591,7 +603,9 @@ Cppyy::TCppType_t Cppyy::GetActualClass(TCppType_t klass, TCppObject_t obj)
     // if the raw name is the empty string (no guarantees that this is so as truly, the
     // address is corrupt, but it is common to be empty), then there is no accessible RTTI
     // and getting the unmangled name will crash ...
-        if (!raw || raw[0] == '\0')
+
+    // a common case are the i/o stream objects b/c these are pulled in from libCling
+        if (!raw || (strstr(cr->GetName(), "std::basic_") && strstr(cr->GetName(), "stream")) || raw[0] == '\0')
             return klass;
     } catch (std::bad_typeid) {
         return klass;        // can't risk passing to ROOT/meta as it may do RTTI
@@ -626,8 +640,16 @@ size_t Cppyy::SizeOf(const std::string& type_name)
 
 bool Cppyy::IsBuiltin(const std::string& type_name)
 {
-    TDataType* dt = gROOT->GetType(TClassEdit::CleanType(type_name.c_str(), 1).c_str());
-    if (dt) return dt->GetType() != kOther_t;
+    if (g_builtins.find(type_name) != g_builtins.end())
+        return true;
+
+    const std::string& tclean = TClassEdit::CleanType(type_name.c_str(), 1);
+    if (g_builtins.find(tclean) != g_builtins.end())
+        return true;
+
+    if (strstr(tclean.c_str(), "std::complex"))
+        return true;
+
     return false;
 }
 
@@ -677,12 +699,13 @@ void Cppyy::Destruct(TCppType_t type, TCppObject_t instance)
     if (cr->ClassProperty() & (kClassHasExplicitDtor | kClassHasImplicitDtor))
         cr->Destructor((void*)instance);
     else {
-        ROOT::DelFunc_t fdel = cr->GetDelete();
+        ::CppyyLegacy::DelFunc_t fdel = cr->GetDelete();
         if (fdel) fdel((void*)instance);
         else {
             auto ib = sHasOperatorDelete.find(type);
             if (ib == sHasOperatorDelete.end()) {
-                sHasOperatorDelete[type] = (bool)cr->GetListOfAllPublicMethods()->FindObject("operator delete");
+                TFunction* f = (TFunction*)cr->GetMethodAllAny("operator delete");
+                sHasOperatorDelete[type] = (bool)(f && (f->Property() & kIsPublic));
                 ib = sHasOperatorDelete.find(type);
             }
             ib->second ? cr->Destructor((void*)instance) : free((void*)instance);
@@ -947,6 +970,15 @@ bool Cppyy::IsEnum(const std::string& type_name)
     return gInterpreter->ClassInfo_IsEnum(tn_short.c_str());
 }
 
+bool Cppyy::IsAggregate(TCppType_t klass)
+{
+// Test if this type is a "plain old data" type
+    TClassRef& cr = type_from_handle(klass);
+    if (cr.GetClass())
+        return cr->ClassProperty() & kClassIsAggregate;
+    return false;
+}
+
 // helpers for stripping scope names
 static
 std::string outer_with_template(const std::string& name)
@@ -1091,10 +1123,10 @@ void Cppyy::GetAllCppNames(TCppScope_t scope, std::set<std::string>& cppnames)
 // add (global) data members
     if (scope == GLOBAL_HANDLE) {
         coll = gROOT->GetListOfGlobals();
-        FILL_COLL(TGlobal, kIsEnum | kIsPrivate | kIsProtected)
+        FILL_COLL(TGlobal, kIsPrivate | kIsProtected)
     } else {
         coll = cr->GetListOfDataMembers();
-        FILL_COLL(TDataMember, kIsEnum | kIsPrivate | kIsProtected)
+        FILL_COLL(TDataMember, kIsPrivate | kIsProtected)
     }
 
 // add enums values only for user classes/namespaces
@@ -1313,11 +1345,14 @@ ptrdiff_t Cppyy::GetBaseOffset(TCppType_t derived, TCppType_t base,
 
 
 // method/function reflection information ------------------------------------
-Cppyy::TCppIndex_t Cppyy::GetNumMethods(TCppScope_t scope)
+Cppyy::TCppIndex_t Cppyy::GetNumMethods(TCppScope_t scope, bool accept_namespace)
 {
-    if (IsNamespace(scope))
+    if (!accept_namespace && IsNamespace(scope))
         return (TCppIndex_t)0;     // enforce lazy
 
+    if (scope == GLOBAL_HANDLE)
+        return gROOT->GetListOfGlobalFunctions(true)->GetSize();
+
     TClassRef& cr = type_from_handle(scope);
     if (cr.GetClass() && cr->GetListOfMethods(true)) {
         Cppyy::TCppIndex_t nMethods = (TCppIndex_t)cr->GetListOfMethods(false)->GetSize();
@@ -1550,9 +1585,12 @@ bool Cppyy::IsConstMethod(TCppMethod_t method)
     return false;
 }
 
-Cppyy::TCppIndex_t Cppyy::GetNumTemplatedMethods(TCppScope_t scope)
+Cppyy::TCppIndex_t Cppyy::GetNumTemplatedMethods(TCppScope_t scope, bool accept_namespace)
 {
-    if (scope == (TCppScope_t)GLOBAL_HANDLE) {
+    if (!accept_namespace && IsNamespace(scope))
+        return (TCppIndex_t)0;     // enforce lazy
+
+    if (scope == GLOBAL_HANDLE) {
         TCollection* coll = gROOT->GetListOfFunctionTemplates();
         if (coll) return (TCppIndex_t)coll->GetSize();
     } else {
@@ -1627,6 +1665,20 @@ bool Cppyy::IsMethodTemplate(TCppScope_t scope, TCppIndex_t idx)
 // helpers for Cppyy::GetMethodTemplate()
 static std::map<TDictionary::DeclId_t, CallWrapper*> gMethodTemplates;
 
+static inline
+void remove_space(std::string& n) {
+   std::string::iterator pos = std::remove_if(n.begin(), n.end(), isspace);
+   n.erase(pos, n.end());
+}
+
+static inline
+bool template_compare(std::string n1, std::string n2) {
+    if (n1.back() == '>') n1 = n1.substr(0, n1.size()-1);
+    remove_space(n1);
+    remove_space(n2);
+    return n2.compare(0, n1.size(), n1) == 0;
+}
+
 Cppyy::TCppMethod_t Cppyy::GetMethodTemplate(
     TCppScope_t scope, const std::string& name, const std::string& proto)
 {
@@ -1644,8 +1696,12 @@ Cppyy::TCppMethod_t Cppyy::GetMethodTemplate(
     TFunction* func = nullptr; ClassInfo_t* cl = nullptr;
     if (scope == (cppyy_scope_t)GLOBAL_HANDLE) {
         func = gROOT->GetGlobalFunctionWithPrototype(name.c_str(), proto.c_str());
-        if (func && name.back() == '>' && name != func->GetName())
-            func = nullptr;  // happens if implicit conversion matches the overload
+        if (func && name.back() == '>') {
+        // make sure that all template parameters match (more are okay, e.g. defaults or
+        // ones derived from the arguments or variadic templates)
+            if (!template_compare(name, func->GetName()))
+                func = nullptr;  // happens if implicit conversion matches the overload
+        }
     } else {
         TClassRef& cr = type_from_handle(scope);
         if (cr.GetClass()) {
@@ -1708,8 +1764,7 @@ Cppyy::TCppMethod_t Cppyy::GetMethodTemplate(
             // allow if requested template names match up to the result
                 const std::string& alt = GetMethodFullName(cppmeth);
                 if (name.size() < alt.size() && alt.find('<') == pos) {
-                    const std::string& partial = name.substr(pos, name.size()-1-pos);
-                    if (strncmp(partial.c_str(), alt.substr(pos, alt.size()-1-pos).c_str(), partial.size()) == 0)
+                    if (template_compare(name, alt))
                         return cppmeth;
                 }
             }
@@ -1720,13 +1775,30 @@ Cppyy::TCppMethod_t Cppyy::GetMethodTemplate(
     return (TCppMethod_t)nullptr;
 }
 
+static inline
+std::string type_remap(const std::string& n1, const std::string& n2)
+{
+// Operator lookups of (C++ string, Python str) should succeeded, for the combos of
+// string/str, wstring/str, string/unicode and wstring/unicode; since C++ does not have a
+// operator+(std::string, std::wstring), we'll have to look up the same type and rely on
+// the converters in CPyCppyy/_cppyy.
+    if (n1 == "str" || n1 == "unicode") {
+        if (n2 == "std::basic_string<wchar_t,std::char_traits<wchar_t>,std::allocator<wchar_t> >")
+            return n2;                      // match like for like
+        return "std::string";               // probably best bet
+    } else if (n1 == "float")
+        return "double";                    // debatable, but probably intended
+    return n1;
+}
+
 Cppyy::TCppIndex_t Cppyy::GetGlobalOperator(
     TCppType_t scope, const std::string& lc, const std::string& rc, const std::string& opname)
 {
 // Find a global operator function with a matching signature; prefer by-ref, but
 // fall back on by-value if that fails.
-    const std::string& lcname = TClassEdit::CleanType(lc.c_str());
-    const std::string& rcname = rc.empty() ? rc : TClassEdit::CleanType(rc.c_str());
+    std::string lcname1 = TClassEdit::CleanType(lc.c_str());
+    const std::string& rcname = rc.empty() ? rc : type_remap(TClassEdit::CleanType(rc.c_str()), lcname1);
+    const std::string& lcname = type_remap(lcname1, rcname);
 
     std::string proto = lcname + "&" + (rc.empty() ? rc : (", " + rcname + "&"));
     if (scope == (cppyy_scope_t)GLOBAL_HANDLE) {
@@ -1797,11 +1869,14 @@ bool Cppyy::IsStaticMethod(TCppMethod_t method)
 }
 
 // data member reflection information ----------------------------------------
-Cppyy::TCppIndex_t Cppyy::GetNumDatamembers(TCppScope_t scope)
+Cppyy::TCppIndex_t Cppyy::GetNumDatamembers(TCppScope_t scope, bool accept_namespace)
 {
-    if (IsNamespace(scope))
+    if (!accept_namespace && IsNamespace(scope))
         return (TCppIndex_t)0;     // enforce lazy
 
+    if (scope == GLOBAL_HANDLE)
+        return gROOT->GetListOfGlobals(true)->GetSize();
+
     TClassRef& cr = type_from_handle(scope);
     if (cr.GetClass() && cr->GetListOfDataMembers())
         return cr->GetListOfDataMembers()->GetSize();
@@ -1839,11 +1914,9 @@ std::string Cppyy::GetDatamemberType(TCppScope_t scope, TCppIndex_t idata)
         TGlobal* gbl = g_globalvars[idata];
         std::string fullType = gbl->GetFullTypeName();
 
-        if ((int)gbl->GetArrayDim() > 1)
-            fullType.append("*");
-        else if ((int)gbl->GetArrayDim() == 1) {
+        if ((int)gbl->GetArrayDim() == 1) {
             std::ostringstream s;
-            s << '[' << gbl->GetMaxIndex(0) << ']' << std::ends;
+            s << '[' << gbl->GetMaxIndex(0) << ']';
             fullType.append(s.str());
         }
         return fullType;
@@ -1863,11 +1936,9 @@ std::string Cppyy::GetDatamemberType(TCppScope_t scope, TCppIndex_t idata)
                 fullType = trueName;
         }
 
-        if ((int)m->GetArrayDim() > 1 || (!m->IsBasic() && m->IsaPointer()))
-            fullType.append("*");
-        else if ((int)m->GetArrayDim() == 1) {
+        if ((int)m->GetArrayDim() == 1) {
             std::ostringstream s;
-            s << '[' << m->GetMaxIndex(0) << ']' << std::ends;
+            s << '[' << m->GetMaxIndex(0) << ']';
             fullType.append(s.str());
         }
         return fullType;
@@ -2108,6 +2179,10 @@ int cppyy_compile(const char* code) {
     return Cppyy::Compile(code);
 }
 
+char* cppyy_to_string(cppyy_type_t klass, cppyy_object_t obj) {
+    return cppstring_to_cstring(Cppyy::ToString(klass, obj));
+}
+
 
 /* name to opaque C++ scope representation -------------------------------- */
 char* cppyy_resolve_name(const char* cppitem_name) {
@@ -2134,6 +2209,14 @@ size_t cppyy_size_of_type(const char* type_name) {
     return Cppyy::SizeOf(type_name);
 }
 
+int cppyy_is_builtin(const char* type_name) {
+    return (int)Cppyy::IsBuiltin(type_name);
+}
+
+int cppyy_is_complete(const char* type_name) {
+    return (int)Cppyy::IsComplete(type_name);
+}
+
 
 /* memory management ------------------------------------------------------ */
 cppyy_object_t cppyy_allocate(cppyy_type_t type) {
@@ -2320,6 +2403,10 @@ int cppyy_is_enum(const char* type_name) {
     return (int)Cppyy::IsEnum(type_name);
 }
 
+int cppyy_is_aggregate(cppyy_type_t type) {
+    return (int)Cppyy::IsAggregate(type);
+}
+
 const char** cppyy_get_all_cpp_names(cppyy_scope_t scope, size_t* count) {
     std::set<std::string> cppnames;
     Cppyy::GetAllCppNames(scope, cppnames);
@@ -2405,6 +2492,10 @@ int cppyy_num_methods(cppyy_scope_t scope) {
     return (int)Cppyy::GetNumMethods(scope);
 }
 
+int cppyy_num_methods_ns(cppyy_scope_t scope) {
+    return (int)Cppyy::GetNumMethods(scope, true);
+}
+
 cppyy_index_t* cppyy_method_indices_from_name(cppyy_scope_t scope, const char* name)
 {
     std::vector<cppyy_index_t> result = Cppyy::GetMethodIndicesFromName(scope, name);
@@ -2472,15 +2563,19 @@ char* cppyy_method_prototype(cppyy_scope_t scope, cppyy_method_t method, int sho
 }
 
 int cppyy_is_const_method(cppyy_method_t method) {
-    return (int)Cppyy::IsConstMethod(method);
+    return (int)Cppyy::IsConstMethod((Cppyy::TCppMethod_t)method);
 }
 
 int cppyy_get_num_templated_methods(cppyy_scope_t scope) {
-    return (int)Cppyy::GetNumTemplatedMethods(scope);
+    return (int)Cppyy::GetNumTemplatedMethods((Cppyy::TCppScope_t)scope);
+}
+
+int cppyy_get_num_templated_methods_ns(cppyy_scope_t scope) {
+    return (int)Cppyy::GetNumTemplatedMethods((Cppyy::TCppScope_t)scope, true);
 }
 
 char* cppyy_get_templated_method_name(cppyy_scope_t scope, cppyy_index_t imeth) {
-    return cppstring_to_cstring(Cppyy::GetTemplatedMethodName(scope, imeth));
+    return cppstring_to_cstring(Cppyy::GetTemplatedMethodName((Cppyy::TCppScope_t)scope, (Cppyy::TCppIndex_t)imeth));
 }
 
 int cppyy_is_templated_constructor(cppyy_scope_t scope, cppyy_index_t imeth) {
@@ -2488,15 +2583,15 @@ int cppyy_is_templated_constructor(cppyy_scope_t scope, cppyy_index_t imeth) {
 }
 
 int cppyy_exists_method_template(cppyy_scope_t scope, const char* name) {
-    return (int)Cppyy::ExistsMethodTemplate(scope, name);
+    return (int)Cppyy::ExistsMethodTemplate((Cppyy::TCppScope_t)scope, name);
 }
 
 int cppyy_method_is_template(cppyy_scope_t scope, cppyy_index_t idx) {
-    return (int)Cppyy::IsMethodTemplate(scope, idx);
+    return (int)Cppyy::IsMethodTemplate((Cppyy::TCppScope_t)scope, idx);
 }
 
 cppyy_method_t cppyy_get_method_template(cppyy_scope_t scope, const char* name, const char* proto) {
-    return cppyy_method_t(Cppyy::GetMethodTemplate(scope, name, proto));
+    return cppyy_method_t(Cppyy::GetMethodTemplate((Cppyy::TCppScope_t)scope, name, proto));
 }
 
 cppyy_index_t cppyy_get_global_operator(cppyy_scope_t scope, cppyy_scope_t lc, cppyy_scope_t rc, const char* op) {
@@ -2531,6 +2626,10 @@ int cppyy_num_datamembers(cppyy_scope_t scope) {
     return (int)Cppyy::GetNumDatamembers(scope);
 }
 
+int cppyy_num_datamembers_ns(cppyy_scope_t scope) {
+    return (int)Cppyy::GetNumDatamembers(scope, true);
+}
+
 char* cppyy_datamember_name(cppyy_scope_t scope, int datamember_index) {
     return cppstring_to_cstring(Cppyy::GetDatamemberName(scope, datamember_index));
 }
@@ -2575,6 +2674,24 @@ int cppyy_get_dimension_size(cppyy_scope_t scope, cppyy_index_t idata, int dimen
 }
 
 
+/* enum properties -------------------------------------------------------- */
+cppyy_enum_t cppyy_get_enum(cppyy_scope_t scope, const char* enum_name) {
+    return Cppyy::GetEnum(scope, enum_name);
+}
+
+cppyy_index_t cppyy_get_num_enum_data(cppyy_enum_t e) {
+    return Cppyy::GetNumEnumData(e);
+}
+
+const char* cppyy_get_enum_data_name(cppyy_enum_t e, cppyy_index_t idata) {
+    return cppstring_to_cstring(Cppyy::GetEnumDataName(e, idata));
+}
+
+long long cppyy_get_enum_data_value(cppyy_enum_t e, cppyy_index_t idata) {
+    return Cppyy::GetEnumDataValue(e, idata);
+}
+
+
 /* misc helpers ----------------------------------------------------------- */
 RPY_EXTERN
 void* cppyy_load_dictionary(const char* lib_name) {
